# vsqtxt2lipsync_02.pl
# 
# extract information from vsq.
#
# usage:
# % mf2t a.vsq  a_vsq.txt
# % perl vsqtxt2lipsync.pl < a_vsq.txt > lipsync.txt
#
# 01. Oct 20, 2007
#   extract text meta events, merge them and chop at \n.
#
# 02. Oct 20, 2007
#   index-to-time conversion implemented.


######################################################################
## default global params.
######################################################################
## smf_type 
##    0: all in one track, 1:track-by-track
## smf_ntracks
##    must be 1 or above
## smf_tpq
##    tick-per-quarter.
##    One quarter note is represented by index interval 480 by default.
$smf_type = 0;       
$smf_ntracks = 0;    
$smf_tpq = 480;

######################################################################
## tempo and timesig 
######################################################################
## We assume that tempo-change / timesig-change will appear in a
## single master-track only, thus they are in the ascending order.
## Therefore sorting is not applied (so far).

## In this tentative version, tempo is assumed to be constant
## throughout the entire data, for simplicity!!

## smf_tempo_list.
## this would look like: ( [idx,tempo,time], [idx,tempo,time], ... )
## to keep the tempo curve.
## Note that this tempo value is "time length of one quarter note in
## microsecond";
## e.x.  666667 microsecond <=> 0.666667 sec for one beat <=> BPM=90.
@smf_tempo_list = ();

## smf_timesig_list.
## this  would look like: ( [idx,meter,num,num], [idx,meter,num,num], .. )
## to keep the timesig markers.
@smf_timesig_list = ();

######################################################################
## track-wise data
######################################################################
$smf_current_track_idx = -1;


######################################################################
## read thru the vsq file (converted to text by mf2t)
######################################################################

while (<>) {
    chop;
    $line = $_;     ### for later use
    @words = split / /;
    if ($words[0] eq "MFile") {
	($dummy, $smf_type, $smf_ntracks, $smf_tpq) = @words;
#	print "smf_type = $smf_type, ";
#	print "smf_ntracks = $smf_ntracks, ";
#	print "smf_tpq = $smf_tpq\n" ;
	next;
    }

    if ($words[1] eq "Tempo") {
	push( @smf_tempo_list, [$words[0], $words[2], 0.0] );
	next;
    }
    if ($words[2] eq "TimeSig") {
	push( @smf_timesig_list, [$words[0], $words[2], $words[3], $words[4]] );
	next;
    }

    if ($words[1] eq "Meta" && $words[2] eq "TrkName") {
	++$smf_current_track_idx;
	my $text = $line;
	$text =~ s/^[0-9]+\s+Meta\s+TrkName\s+//;
	$text =~ s/^\"//;
	$text =~ s/\"$//;
	$smf_track_name[$smf_current_track_idx] = $text;
    }
    if ($smf_current_track_idx < 0) { next; }

    ### the following section is visited only after 
    ### the first "TrkName" is reached (that is, smf_current_track_idx
    ### has been set to 0)

    if ( ($words[1] eq "Meta") && ($words[2] eq "Text") && ($words[3] =~ /^\"DM/)) {
	$line =~ s/^[0-9]+\s+Meta\s+Text\s+//;
	$line =~ s/^\"DM:[0-9]{4}://;
	$line =~ s/\"$//;
	$smf_metatext_dm[$smf_current_track_idx] .= $line;
    }
}


########################################
## post processing : generate index-time map
########################################
$time_now_usec = 0.0;
$prev_idx = 0;
$prev_tempo = 500000;
foreach $pIdxTempo ( @smf_tempo_list ) {
    my ($idx, $tempo) = @$pIdxTempo;
    $time_now_usec += $prev_tempo * ($idx - $prev_idx) / $smf_tpq;
    $prev_idx = $idx;
    $prev_tempo = $tempo;
    $pIdxTempo->[2] = $time_now_usec;
#    print "$pIdxTempo->[0]   $pIdxTempo->[1]   $pIdxTempo->[2]\n";
}    
## after this processing, smf_tempo_list looks like:
##   ( [idx,tempo, time], [idx,tempo, time], [idx,tempo, time].... )


######################################################################
## conversion func. from sequence index to absolute time (in micro-sec).
######################################################################
## 
## Note: you might think that binary-search would make things faster;
## However, given the fact that the number of tempo-changes are at
## maximum 10 or so, simple linear-search would not be so bad.

sub idx_to_time {
    my ( $target ) = @_;
    my ( $pIdxTempo, $idx, $tempo, $timeoffset );
    foreach $pIdxTempo (reverse @smf_tempo_list) {
#	printf( "target=%d,  table=%d\n", $target, $pIdxTempo->[0] );
	if ( $target >= ($pIdxTempo->[0]) ) {
	    ($idx, $tempo, $timeoffset) = @$pIdxTempo;
	    last;
	}
    }
#    printf("point chosen: idx=%d, tempo=%d, time=%f\n", $idx, $tempo, $timeoffset );
    return (0.0 + $target - $idx) * $tempo / $smf_tpq + $timeoffset;
}

# testing code for idx_to_time.
# I had manually inserted the following lines to miku_vsq.txt, after
# the line that said "Tempo 666667" for test.
#700 Tempo 500000
#1000 Tempo 800000
#2000 Tempo 750000
#3000 Tempo 500000
#for $i ( 0, 1, 500, 999, 1000, 1001, 1199, 1200, 1999, 2000, 2001 ) {
#    $t = idx_to_time($i);
#    printf( "%5d => %f\n", $i, $t );
#}


########################################
### fix merged texts  and store in $smf_metatext_dm[trk].
########################################

for ( $i = 0; $i <= $smf_current_track_idx; ++$i ) {
    print "***($smf_track_name[$i])***\n";
    my $vsq_text = $smf_metatext_dm[$i];
    $vsq_text =~ s/\\n/\n/g;
    $vsq_text =~ s/\\\"/\"/g;
    $smf_vsq_text[$i] = $vsq_text; # save a string of all the text

#    test dump.
#    print $vsq_text;

## read in the text into a structure
    my $ref = vsq_text_to_structure( $vsq_text );

## print the structure in the text format, for test
##    vsq_structure_to_text( $ref );

    vsq_structure_print_events( $ref ); 
}

######################################################################
## read-in vsq_text  and store it into a structure, of which reference
## is returned.
##
## pRef= vsq_text_to_structure($text);
##
## print $pRef->{"Common"}->{"Version"};
## .... this will print "DSB301", or something like that.
## 
## section "EventList", which associates tick indice with ID#nnnn, 
## is stored as "timevalue" type, which can be accessed as
## print $pRef->{"EventList"}->[0]->[0]
## .... the first tick index .
## print $pRef->{"EventList"}->[0]->[1]
## .... "ID#nnnn"  for the first tick index.
## 
## section "*BPList", which associates tick indice to a certain value,
## is stored as "timevalue" type, which can be accessed as
## print $pRef->{"XXXBPList"}->[0]->[0]
## .... the first tick index.
## print $pRef->{"XXXBPList"}->[0]->[1]
## .... number  for the first tick index.
##
## To know if the set is a timevalue or an array, 
## the following is used.
## if ( $pRef->{"isTimevalue"}->{$section} == 1 ) { ...
##
## The order of the sections is preserved in a list at:
##  $pRef->{"sectionOrder"}

sub vsq_text_to_structure {
    my ($vsq_text) = @_;
    my $pVsqStructure = {};
    my ( $section, $pNewSet, $isTimevalue );
    chop $vsq_text;
    my $pIsTimevalue = {};
    my $pSectionOrder = [];
    $pVsqStructure->{"isTimevalue"} = $pIsTimevalue;
    $pVsqStructure->{"sectionOrder"} = $pSectionOrder;
    foreach $line (split( /\n/, $vsq_text )) {
	if ($line =~ /^\[(.+)\]$/ ) {
	    $section=$1;
	    if ($section =~ /BPList$/ ||$section eq "EventList" ) {
		$pNewSet = [];
		$isTimevalue = 1;
	    } else {
		$pNewSet = {};
		$isTimevalue = 0;
	    }
	    $pVsqStructure->{$section} = $pNewSet;
	    $pIsTimevalue->{$section} = $isTimevalue;
	    push( @$pSectionOrder, $section );
	    next;
	}

	$line =~ /(^[^=]+)=(.*)$/;
	if ( $isTimevalue ) {
	    push (@$pNewSet, [$1, $2] );
	} else {
	    $pNewSet->{$1} = $2;
	}
    }
    return $pVsqStructure;
}


######################################################################
## print vsq_structure
######################################################################

sub vsq_structure_to_text {
    my ($ref) = @_;

    my ($pair, $pSection, $section, $key);
    
    my $pSectionList = $ref->{"sectionOrder"};

    foreach $section (@$pSectionList) {
	printf ("[%s]\n", $section );
	$pSection = $ref->{$section};
	if ( $ref->{"isTimevalue"}->{$section} == 1 ) {
	    foreach $pair ( @$pSection ) {
		printf("%d=%s\n", $pair->[0], $pair->[1] );
	    }
	} else {
	    foreach $key (keys %$pSection) {
		printf("%s=%s\n", $key, $pSection->{$key} );
	    }
	}
    }
}


######################################################################
## convert notenumber to note-name
######################################################################
# 23->-1, 24->0 35->0 36->1   59->2, 60->3, 61->3, 71->3, 72->4
sub note_number_to_name {
    my ($num) = @_;
    my @pitchclass = ("C","C#", "D", "D#","E", "F","F#","G","G#","A","A#","B");
    my ($oct) = int(($num-24) / 12 );
    my ($name) = $pitchclass[$num % 12];
    return sprintf("%-2s %2d", $name,$oct);
}    

######################################################################
## extract events from the structure and print them out 
######################################################################
sub vsq_structure_print_events {
    my ($ref) = @_;
    my $pEventList = $ref->{"EventList"};  # this is a timevalue set.
    my $pPair;
    my $end_time;
    my $prev_end_time = -1.0;
    foreach $pPair (@$pEventList) {
	my ($begin_tick_index, $idnum) = @$pPair;
	if ( $ref->{$idnum}->{"Type"} ne "Anote" ) { next; }

	my $end_tick_index = $begin_tick_index + $ref->{$idnum}->{"Length"};
	my $dynamics = $ref->{$idnum}->{"Dynamics"};
	my $notenum = $ref->{$idnum}->{"Note#"};
	my $notename = &note_number_to_name( $notenum );
	my $lyricHandle = $ref->{$idnum}->{"LyricHandle"};
	my ($chars_hex, $phonetic, $v0, $v1, $v2, $v3) = split( /,/, $ref->{$lyricHandle}->{"L0"} );
	$chars_hex =~ s/[\"\\x]//g;
	my $chars = pack "H*", $chars_hex;
	$phonetic =~ s/\"//g;
	my $beg_time = &idx_to_time($begin_tick_index)/1000000.0;
	$end_time = &idx_to_time($end_tick_index)/1000000.0;
	
	if ( $prev_end_time >= 0.0 ) {
	    if ( $beg_time < $prev_end_time ) {
		$prev_end_time = $beg_time;
	    }
	    printf( "%10.3f\t%s\n", $prev_end_time, "end" );
	}
	$prev_end_time = $end_time;
	printf( "%10.3f\t%s\t%s\t%s\t%d\t%d\t%s\n", $beg_time, "beg", $chars, $phonetic, $dynamics, $notenum, $notename );
    }
    printf( "%10.3f\t%s\n", $end_time, "end" );
}
